\ FLOATING and FLOATING-EXT words not implemented here are:
\ FLITERAL REPRESENT F** FACOS FACOSH FALOG FASIN FASINH FATAN FATAN2
\ FATANH FCOS FCOSH FEXP FEXPM1 FLN FLNP1 FLOG FSIN FSINCOS FSINH FTAN
\ and FTANH

\ 0: Initial release 11/07/2002, Brad Eckert 
\ 1: Removed locals from F*, PAD from F/. Added FSQRT.
\ 2: Fixed FROUND, F.
\ 3: Fixed FPICK, converted standard words to upper case, sped up stack.
\ 4: Limited number of significant digits in F. to comply with ANS standard.
\ 5: Changed F2* and F2/ to use fshift. You can use fshift for quick *2^n.
\    Replaced DO..LOOP with 0 ?DO LOOP structure.
\ 6: Changed 0 ?DO LOOP to use EVALUATE (more optimizer-friendly).
\ 7: Fixed a bug in F/.
\ 8: Removed extra FDUPs from F<. Fixed a bug in F+. Added >FLOAT FS. FE. F.S
\ 9: Corrected (F.) rounding error and added more tests.
\ 10: Changed -ROT to ROT ROT for strict ANS compliance.
\ 11: Made F# state smart. Fixed bug in FSQRT.
\ 12: Made 4tH version, Hans Bezemer
\ 13: Fixed FLOOR, added F=
\ 14: Added IEEE 754 exceptions
\ 15: Fixed FSQRT (didn't clear stack on error)

[UNDEFINED] floats [IF]
[UNDEFINED] 2@   [IF] include lib/anscore.4th  [THEN]
[UNDEFINED] um*  [IF] include lib/mixed.4th    [THEN]
[UNDEFINED] 2rot [IF] include lib/2rotover.4th [THEN]
[UNDEFINED] pick [IF] include lib/pickroll.4th [THEN]

[UNDEFINED] FLOATING-STACK [IF]
8 CONSTANT FLOATING-STACK              \ size of floating point stack
[THEN]

\ Floating point representation uses three cells:
\ 0 upper mantissa
\ 1 lower mantissa
\ 2 exponent: MSB = sign of mantissa, others = signed exponent

3 CELLS CONSTANT c/float
c/float [NEGATE] CONSTANT -c/float
c/float CONSTANT FLOAT

VARIABLE sigdigits                     \ # of digits to display (see F., R.)
VARIABLE fsp                           \ stack pointer

FLOAT FLOATING-STACK [*] ARRAY fstak   \ floating point stack
FLOAT 2 [*] ARRAY ftemp                \ temporary float variables

/cell 8 [*] CONSTANT bits/cell         \ 16-bit or 32-bit Forth

[hex]
max-n    CONSTANT &unsign
(error)  CONSTANT &sign 
40000000 CONSTANT &esign               \ 40...
[decimal]

\ A variable called ferror stores error codes in the event of an error.
\ Execution is not interrupted, but a bad result could be generated. You
\ should check for a non-zero error code at appropriate times.

VARIABLE ferror                        \ last seen error

\ Standard IEEE 754 exceptions

0   enum FE.NOERRORS                   \ No FP errors
    enum FE.DIVBYZERO                  \ Division by zero
    enum FE.OVERFLOW                   \ Overflow
    enum FE.UNDERFLOW                  \ Underflow
    enum FE.INVALID                    \ Invalid operation
constant FE.INEXACT                    \ Inexact result

: ud2/          ( d -- d' )
                D2/ &unsign AND ;

: frshift       ( count 'man -- )      \ right shift mantissa
                SWAP 0 MAX bits/cell 2* MIN
                >R DUP 2@ R> 0 ?DO ud2/ LOOP ROT 2! ;

: exp>sign      ( exp -- exp' sign )
                DUP &unsign AND        \ mask off exponent
                DUP &esign AND 2* OR   \ sign extend exponent
                SWAP &sign AND ;       \ get sign of mantissa

: sign>exp      ( exp sign -- exp' )
                SWAP &unsign AND OR ;

: t2*           ( triple -- triple' )
                D2* ROT DUP 0< 1 AND >R 2* ROT ROT R> 0 D+ ;

: t2/           ( triple -- triple' )
                OVER 1 AND 0 SWAP IF INVERT THEN &sign AND >R D2/ ROT 
                1 RSHIFT R> OR ROT ROT ;
                                       \ MHL|C
: t+            ( t1 t2 -- t3 )
                2>R >R ROT 0 R> 0 D+ 0 2R> D+
                ROT >R D+ R> ROT ROT ;

: tneg          ( d flag -- t )        \ conditionally negate d to 3-cell t
                0<> >R 2DUP OR 0<> R> AND IF DNEGATE -1 ELSE 0 THEN ;                

: lstemp        ( -- )                 \ 6-cell left shift FTEMP
                ftemp 6 CELLS + 0      ( *LSB carry . . )
      6 0 ?DO   >R -1 CELLS +          \ LSB in high memory
                DUP @ 0 D2* SWAP R> +  ( a co n ) \ ROL
                ROT TUCK ! SWAP        \ a carry
        LOOP    2DROP ;

: *norm         ( triple exp -- double exp' )         \ normalize triple
     >R BEGIN   DUP 0< 0=
        WHILE   t2* R> 1- >R
        REPEAT  &sign 0 0 t+           \ round
                ROT DROP R> ;

: longdiv       ( DA DB -- DA/DB )     \ fractional divide
                0 0 ftemp 2!
                bits/cell 2* 1+        \ long division
        0 ?DO   2OVER 2OVER DU<        \ double/double
                IF   0
                ELSE 2DUP 2>R D- 2R> 1 \ a b --> a-b b
                THEN 0 ftemp 2@ D2* D+ ftemp 2!
                2SWAP D2* 2SWAP
        LOOP    DU< 0= 1 AND 0         \ round
                ftemp 2@ D+ ;

\ The rest can be left in high level Forth...

: 'm1           ( -- addr )      fsp @ 3 cells - ;    \ -> 1st stack mantissa
: 'm2           ( -- addr )      fsp @ 6 cells - ;    \ -> 2nd stack mantissa
: 'm3           ( -- addr )      fsp @ 9 cells - ;    \ -> 3rd stack mantissa
: 'e1           ( -- addr )      fsp @ 1 cells - ;    \ -> 1st stack exponent
: 'e2           ( -- addr )      fsp @ 4 cells - ;    \ -> 2nd stack exponent
: fmove         ( a1 a2 -- )     c/float 0 DO OVER I + @ OVER I + ! LOOP 2DROP ;
: m1get         ( addr -- )      'm1 fmove ;          \ move to M1
: m1sto         ( addr -- )      'm1 SWAP fmove ;     \ move from M1
: expx1         ( -- exp sign )  'e1 @ exp>sign ;
: expx2         ( -- exp sign )  'e2 @ exp>sign ;
: ftemp'        ( -- addr )      ftemp 2 CELLS + ;
: >exp1         ( exp sign -- )  sign>exp 'e1 ! ;
: fshift        ( n -- )         expx1 >R + R> >exp1 ; \ F = F * 2^n

\ A normalized float has an unsigned mantissa with its MSB = 1

: normalize     ( -- )
                'm1 2@ 2DUP OR 0=
       IF       2DROP                  \ Zero is a special case.
       ELSE     0 ROT ROT expx1 >R *norm
                R> >exp1 'm1 2!
       THEN     ;

\ *S Floating Point Words

\ *+
: F2*           ( fs: r1 -- r3 )  1 fshift ;
: F2/           ( fs: r1 -- r3 ) -1 fshift ;
: PRECISION     ( -- n )         sigdigits @ ;  \ max decimal digits/double
: SET-PRECISION ( n -- ) bits/cell -1 [+] 602 [*] 1000 / MIN sigdigits ! ;
: FCLEAR        ( -- )           fstak fsp ! FE.NOERRORS ferror ! ;
: FDEPTH        ( -- )           fsp @ fstak - c/float / ;
: FDUP          ( fs: r -- r r ) c/float fsp +! 'm2 m1get ;
: FDROP         ( fs: r -- )    -c/float fsp +! ;
: FNEGATE       ( fs: r1 -- r3 ) 'e1 @ &sign XOR 'e1 ! ;
: D>F           ( d -- | -- r )  FDUP DUP 0< IF DNEGATE &sign ELSE 0 THEN 
                                 'e1 ! 'm1 2!  normalize ;
: F>D           ( -- d | r -- )  expx1 >R DUP 1- 0<
        IF      NEGATE &unsign AND 'm1 frshift 'm1 2@ R> IF DNEGATE THEN
        ELSE    R> 2DROP -1 &unsign FE.OVERFLOW ferror !
        THEN    FDROP ;                \ overflow: return maxdint
: S>F           ( n -- | -- r )  DUP ABS U>D ROT 0< IF DNEGATE THEN D>F ;
: FLOAT+        ( n1 -- n2 )     c/float + ;
: FLOATS        ( n1 -- n2 )     c/float * ;
: F@            ( a -- | -- r )  FDUP m1get ;
: F!            ( a -- | r -- )  m1sto FDROP ;
: F0=           ( -- t | r -- )  'm1 2@ OR 0= FDROP ;
: F0<           ( -- t | r -- )  'e1 @ 0< FDROP ;
: FLOOR         FDUP F>D F0< IF 1 U>D DNEGATE D+ THEN D>F ;
: FABS          ( fs: r1 -- r3 ) 'e1 @ 0< IF FNEGATE THEN ;
: FALIGN        ( -- )           ALIGN ;
: FALIGNED      ( addr -- addr ) ALIGNED ;
: FSWAP         ( fs: r1 r2 -- r2 r1 )
                'm1 ftemp fmove 'm2 m1get ftemp 'm2 fmove ;
: FOVER         ( fs: r1 r2 -- r1 r2 r3 )
                c/float fsp +! 'm3 m1get ;
: FPICK         ( n -- ) ( fs: -- r )
                c/float fsp +! 2 + -c/float * fsp @ + m1get ;
: FNIP          ( fs: r1 r2 -- r2 ) FSWAP FDROP ;
: FROT          ( fs: r1 r2 r3 -- r2 r3 r1 )
                'm3 ftemp fmove
                'm2 'm3 c/float 2* 0 DO OVER I + @ OVER I + ! LOOP 2DROP
                ftemp m1get ;
: F+            ( fs: r1 r2 -- r3 )    \ Add two floats
       FDUP F0= IF FDROP EXIT THEN                    \ r2 = 0
      FOVER F0= IF FNIP 'e1 @ 0< IF FNEGATE THEN EXIT THEN
                expx1 >R expx2 >R -  DUP 0<           \ r1 = 0
        IF      NEGATE 'm1 frshift 0   \ align exponents
        ELSE    DUP    'm2 frshift
        THEN    'e2 @ +
                'm2 2@ R> tneg
                'm1 2@ R> tneg
                t+ DUP >R              ( exp L M H | sign )
    DUP IF      t2/ IF DNEGATE THEN 'm2 2! 1+
        ELSE    DROP 'm2 2!
        THEN    R> &sign AND sign>exp 'e2 !
                FDROP normalize ;

: F-            ( fs: r1 r2 -- r3 )    \ Subtract two floats
                FNEGATE F+ ;

: F<            ( -- flag ) ( F: r1 r2 -- )           \ Compare two floats
                F- F0< ;

: F=            ( -- flag ) ( F: r1 r2 -- )           \ Compare two floats
                F- F0= ;

: FROUND        ( fs: r1 -- r2 )       \ Round to the nearest integer
                expx1 >R NEGATE 1- 'm1 frshift        \ convert to half steps
                'm1 2@ 1 U>D D+ SWAP -2 AND SWAP      \ round
                'm1 2! -1 R> >exp1 normalize ;        \ re-float

: FMIN          ( F: r1 r2 -- rmin ) FOVER FOVER F<
                IF FDROP ELSE FNIP THEN ;

: FMAX          ( F: r1 r2 -- rmax ) FOVER FOVER F<
                IF FNIP ELSE FDROP THEN ;

: F*            ( fs: r1 r2 -- r3 )    \ Multiply two floats
                'm1 2@ 'm2 2@
                OVER >R ftemp' 2!
                OVER >R ftemp  2!
                R> R> OR 
                                       \ need long multiply?
        IF      FTEMP CELL+ @ FTEMP' CELL+ @ UM* &sign 0 D+ NIP \ round
                FTEMP @       FTEMP' @       UM*
                FTEMP CELL+ @ FTEMP' @       UM* 0 t+
                FTEMP @       FTEMP' CELL+ @ UM* 0 t+
        ELSE    0 ftemp @ ftemp' @ UM* \ lower parts are 0
        THEN    2DUP OR 3 PICK OR      \ zero?
        IF      expx1 >R expx2 >R + bits/cell 2* + *norm
                R> R> XOR sign>exp 'e2 !
        ELSE    DROP                   \ zero result
        THEN    'm2 2! FDROP ;

: F/            ( fs: r1 r2 -- r3 )    \ Divide r1 by r2
                FDUP F0=                              \ div by 0, man= umaxint
        IF      FDROP -1 -1 'm1 2! FE.DIVBYZERO ferror !
                'e1 @ &sign AND &esign 1- OR 'e1 !    \   exponent = maxint/2
        ELSE    'm2 2@ 'm1 2@
                DU< 0= IF 1 'm2 frshift F2/ THEN      \ divisor<=dividend
                'm1 CELL+ @
           IF   'm2 2@ ud2/ 'm1 2@ ud2/  longdiv      \ max divisor = dmaxint
           ELSE 0 'm2 2@ 'm1 @ DUP >R UM/MOD          \ 0 rem quot | divisor
                ROT ROT R@ UM/MOD ROT ROT R> 1 RSHIFT &unsign AND U>
                IF 1 U>D D+ THEN                      \ round
           THEN expx2 >R expx1 >R - bits/cell 2* -
                >R 'm2 2! R> R> R> XOR sign>exp 'E2 !
                FDROP
        THEN    ;

: F~            ( f: r1 r2 r3 -- ) ( -- flag )        \ f-proximate
                FDUP F0<                              \ Win32forth version
        IF      FABS FOVER FABS 3 FPICK FABS F+ F*    \ r1 r2 r3*(r1+r2)
                FROT FROT F- FABS FSWAP F<
        ELSE    FDUP F0=
                IF      FDROP F- F0=
                ELSE    FROT FROT F- FABS FSWAP F<
                THEN
        THEN ;

: FSQRT         ( fs: r -- r' )        \ Square root
                expx1 IF drop FE.INVALID ferror ! EXIT THEN \ sqrt of negative
                DUP 1 AND DUP >R +     \ round up exponent/2
                2/ bits/cell - 0 >exp1
                ftemp 6 BOUNDS DO 0 I ! LOOP
                'm1 2@                 \ x
                R> IF ud2/ THEN        \ exponent was rounded
                ftemp 2 CELLS + 2!     \ x*2^(2*bits/cell)
                0 0  bits/cell 2*      \ p = 0
        0 ?DO   lstemp lstemp          \ shift left x into a 2 places
                D2*                    \ shift left p one place
                2DUP D2* ftemp 2@ D<
                IF      ftemp 2@ 2OVER D2* 1 U>D D+ D-
                        ftemp 2!       \ a:=a-(2*p+1)
                        1 U>D D+       \ p:=p+1
                THEN
        LOOP    'm1 2! normalize ;

\ For fixed-width fields, it makes sense to use fsplit and (F.) for output.
\ fsplit converts a float to integers suitable for pictured numeric output.
\ Fracdigits is the number of digits to the right of the decimal.

: fsplit        ( F: r -- ) ( fracdigits -- sign Dint Dfrac )
\ Split float into integer component parts.
                >R expx1 NIP FABS      \ int part must fit in a double
                FDUP F>D 2DUP D>F F-   \ get int, leave frac
            2 0 R> 0 ?DO D2* 2DUP D2* D2* D+ LOOP     \ 2 * 10^precision
                D>F F* F>D 1 U>D D+ ud2/ ;            \ round

64 string floathold
variable hld

: num>char dup 9 > if 7 + then [char] 0 + ;
: fhold 1 hld +! floathold hld @ - c! ;
: f# base @ mu/mod rot num>char fhold ;

\ (F.) uses PRECISION as the number of digits after the decimal. F. clips off
\ the result to avoid displaying extra (possibly garbage) digits. However,
\ this defeats last-digit rounding. (F.) TYPE is the prefered display method.
\ F. is included for completeness.

: (F.)          ( F: r -- ) ( -- addr len )
\ Convert float to a string
                0 hld ! FDEPTH 1- 0< IF 0 0 EXIT THEN \ empty stack -> blank
                PRECISION fsplit
                PRECISION 0 ?DO f# LOOP D+
                PRECISION IF [CHAR] . fhold THEN
                BEGIN f# 2DUP OR 0= UNTIL
                ROT 0< IF [CHAR] - fhold THEN
                2DROP hld @ floathold OVER - SWAP ;

: F.            ( F: r -- )  (F.) PRECISION 1+ MIN TYPE SPACE ;
: R.            ( F: r -- )  (F.) TYPE SPACE ;

: (E.)          ( stepsize resolution -- | F: r -- )  \ X.XXXXXXEYY format
                >R FDUP FABS 0                  ( step 0 )
        BEGIN   FDUP 1 S>F F<
        WHILE   OVER - R@ S>F F*
        REPEAT
        BEGIN   FDUP R@ S>F F< 0=
        WHILE   OVER + R@ S>F F/
        REPEAT  R> DROP NIP
                FSWAP F0< IF FNEGATE THEN
                (F.) TYPE
                DUP ABS S>D <# #S SIGN
                [CHAR] E HOLD #> TYPE SPACE ;

: FS.   ( F: r -- ) 1   10 (E.) ;      \ scientific notation
: FE.   ( F: r -- ) 3 1000 (E.) ;      \ engineering notation

\ String to floating point conversion ---------------------------------------

: fsign         ( a n -- a' n' sign )  \ get sign from the input string
                OVER C@ OVER IF DUP
                  [CHAR] - = IF DROP CHOP -1 EXIT ELSE
                  [CHAR] + = IF CHOP THEN THEN
                ELSE DROP THEN 0
;

variable flgood

: fdigit?       ( a len -- a len n f ) \ get digit from the input string
                DUP 0<> >R
                OVER C@ [CHAR] 0 - DUP 0< OVER 9 > OR 0=
                R> AND  DUP
        IF      2SWAP CHOP 2SWAP       \ good digit, use it
                1 flgood +!
        THEN    ;

: flint         ( addr len -- addr' len' )
        BEGIN   fdigit?                \ get integer part
        WHILE   10 S>F F* S>F F+
        REPEAT  DROP ;

: flexp         ( addr len -- addr' len' )            \ get exponent
                OVER C@ [CHAR] D -     \ expect 0,1,20,21
                -34 AND 0=             \ and invert(0x21)
                flgood @ 1 <> AND      \ some mantissa digits are required
        IF      CHOP fsign >R 0 >R     \ E,e,D,d is valid
         BEGIN  fdigit?                \ get exponent
         WHILE  R> 10 * + >R
         REPEAT DROP R> R>
         IF     0 ?DO 10 S>F F/ LOOP   \ positive exponent
         ELSE   0 ?DO 10 S>F F* LOOP   \ negative exponent
         THEN
        THEN    DUP
        IF      0 flgood !             \ unknown text left to convert
        THEN    ;

: flfrac        ( addr len -- addr' len' )
                CHOP 1 S>F             \ convert after the decimal
        BEGIN   fdigit?
        WHILE   10 S>F F/              ( F: num digitmul )
                FDUP S>F F*            ( F: num digitmul delta )
                FROT F+ FSWAP
        REPEAT  FDROP DROP DUP         \ more to convert?
        IF      flexp
        THEN    ;

: >FLOAT        ( addr len -- f ) ( f: -- r | <nothing> )
\ Convert base 10 string to float regardless of BASE.
                -TRAILING  0 S>F
                fsign >R   1 flgood !
                flint DUP
        IF      OVER C@ [CHAR] . =
                IF flfrac ELSE flexp THEN
        THEN    2DROP
                R> IF FNEGATE THEN
                flgood @ IF true ELSE FDROP false THEN
                ;

: S>FLOAT >FLOAT 0= ABORT" Bad float" ;

: F.S ( -- )   FDEPTH BEGIN ?DUP WHILE 1- DUP FPICK F. REPEAT CR ;
: PI  ( r -- ) s" 3.1415926535897932384626433832795" s>float ;
: E   ( r -- ) s" 2.7182818284590452353602874713527" s>float ;

\ test goodies
false [IF]
fclear
100 SET-PRECISION                      \ max precision for testing

CR .( Basic arithmetic ------------)
CR .( 1/7 = ) 1 S>F 7 S>F F/ R.
CR .( 1/3 = ) 1 S>F 3 S>F F/ R.
CR .( 2/3 = ) 2 S>F 3 S>F F/ R.
CR .( 355/113 = ) 355 S>F 113 S>F F/ R.
CR .( sqrt[2] = ) 2 S>F FSQRT R.

6 SET-PRECISION

CR .( Numeric conversion ----------)
CR .( .0001 =         ) 1 S>F 10000 S>F F/ FE.
CR .( -.0001 =        ) s" -.0001" s>float FE.
CR .( 10000 * .0001 = ) s" 10000" s>float s" .0001" s>float F* F.
CR .( 1M, 1u =        ) s" 1d6" s>float  FS. s" 1D-6" s>float FS.
CR .( -1.23M, 1.23u = ) s" -1.23e6" s>float FS. s" 1.23e-6" s>float FS.
CR .( sqrt[1e99] =    ) s" 1e99" s>float FSQRT FS.
CR .( 123 + 456 =     ) s" 123" s>float s" 456" s>float F+ F.
CR .( 123 - 456 =     ) s" 123" s>float s" 456" s>float F- F.
CR .( 456 - 123 =     ) s" 456" s>float s" 123" s>float F- F.
[THEN]

[DEFINED] 4TH# [IF]
hide c/float
hide -c/float
hide sigdigits
hide fsp
hide fstak
hide ftemp
hide bits/cell
hide &unsign
hide &sign
hide &esign
hide ud2/
hide frshift
hide exp>sign
hide sign>exp
hide t2*
hide t2/
hide t+
hide tneg
hide lstemp
hide *norm
hide longdiv
hide 'm1
hide 'm2
hide 'm3
hide 'e1
hide 'e2
hide fmove
hide m1get
hide m1sto
hide expx1
hide expx2
hide ftemp'
hide >exp1
hide fshift
hide normalize
hide fsplit
hide floathold
hide hld
hide num>char
hide fhold
hide f#
hide fsign
hide flgood
hide fdigit?
hide flint
hide flexp
hide flfrac
[THEN]
[THEN]
